import * as tsplus_module_1 from "@tsplus/stdlib/data/Function";
import * as tsplus_module_2 from "@tsplus/stdlib/structure/Equals";
export const zipWith = zipWith_1;
export const map = map_1;
export const flatMap = flatMap_1;
export const concat = concat_1;
export const prepend = prepend_1;
export const prependOperator = prependOperator_1;
export const append = append_1;
export const equalsWith = equalsWith_1;

function* genOf(a) {
  yield a;
}

function* genMap(iterator, mapping) {
  let n = -1;

  while (true) {
    const result = iterator.next();

    if (result.done) {
      break;
    }

    n += 1;
    yield mapping(result.value, n);
  }
}

function* genChain(iterator, mapping) {
  while (true) {
    const result = iterator.next();

    if (result.done) {
      break;
    }

    const ib = mapping(result.value)[Symbol.iterator]();

    while (true) {
      const result = ib.next();

      if (result.done) {
        break;
      }

      yield result.value;
    }
  }
}
/**
 * Zips the values of both iterators with the provided zipper function
 *
 * @tsplus static Collection.Aspects zipWith
 * @tsplus pipeable Collection zipWith
 */


function zipWith_1(that, zipper) {
  return self => {
    // inspired from "Closing Iterables is a Leaky Abstraction" by Reginald Braithwaite
    // https://raganwald.com/2017/07/22/closing-iterables-is-a-leaky-abstraction.html
    return {
      [Symbol.iterator]() {
        let done = false;
        const ia = self[Symbol.iterator]();
        const ib = that[Symbol.iterator]();
        return {
          next() {
            if (done) {
              // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
              return this.return();
            }

            const va = ia.next();
            const vb = ib.next();
            return va.done || vb.done ? // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            this.return() : {
              done: false,
              value: zipper(va.value, vb.value)
            };
          },

          return(value) {
            if (!done) {
              done = true;

              if (typeof ia.return === "function") {
                ia.return();
              }

              if (typeof ib.return === "function") {
                ib.return();
              }
            }

            return {
              done: true,
              value
            };
          }

        };
      }

    };
  };
}
/**
 * Maps the values of the iterator using the provided function
 *
 * @tsplus static Collection.Aspects map
 * @tsplus pipeable Collection map
 */


function map_1(f) {
  return self => {
    return {
      [Symbol.iterator]: () => genMap(self[Symbol.iterator](), f)
    };
  };
}
/**
 * Zips the two iterators into an iterator of a tuple
 *
 * @tsplus static Collection.Aspects zip
 * @tsplus pipeable Collection zip
 */


export function zip(that) {
  return self => zipWith_1(that, (a, b) => [a, b])(self);
}
/**
 * Maps the iterator using the provided function and flatten its result
 *
 * @tsplus static Collection.Aspects flatMap
 * @tsplus pipeable Collection flatMap
 */

function flatMap_1(f) {
  return self => {
    return {
      [Symbol.iterator]: () => genChain(self[Symbol.iterator](), f)
    };
  };
}
/**
 * Applicative's apply
 *
 * @tsplus static Collection.Aspects ap
 * @tsplus pipeable Collection ap
 */


export function ap(fa) {
  return fab => flatMap_1(f => map_1(f)(fa))(fab);
}
/**
 * Creates an iterator of a single value
 *
 * @tsplus static Collection.Ops of
 */

export function of(a) {
  return {
    [Symbol.iterator]: () => genOf(a)
  };
}
/**
 * Takes the fist n elements
 *
 * @tsplus static Collection.Aspects take
 * @tsplus pipeable Collection take
 */

export function take(n) {
  return self => {
    return {
      *[Symbol.iterator]() {
        let i = 0;

        for (const x of self) {
          if (i++ >= n) {
            return;
          }

          yield x;
        }
      }

    };
  };
}
/**
 * Skips the first n elements
 *
 * @tsplus static Collection.Aspects skip
 * @tsplus pipeable Collection skip
 */

export function skip(n) {
  return a => {
    return {
      *[Symbol.iterator]() {
        let i = 0;

        for (const x of a) {
          if (i++ >= n) {
            yield x;
          }
        }
      }

    };
  };
}
/**
 * Empty iterator
 *
 * @tsplus static Collection.Ops never
 */

export const never = {
  // eslint-disable-next-line @typescript-eslint/no-empty-function
  *[Symbol.iterator]() {}

};
/**
 * Loops over the iterator accumulating a result using the provided function giving access to the element index
 *
 * @tsplus static Collection.Aspects reduceWithIndex
 * @tsplus pipeable Collection reduceWithIndex
 */

export function reduceWithIndex(b, f) {
  return self => {
    let res = b;
    let n = -1;
    const iterator = self[Symbol.iterator](); // eslint-disable-next-line no-constant-condition

    while (true) {
      const result = iterator.next();

      if (result.done) {
        break;
      }

      n += 1;
      res = f(res, result.value, n);
    }

    return res;
  };
}
/**
 * Loops over the iterator accumulating a result using the provided function
 *
 * @tsplus static Collection.Aspects reduce
 * @tsplus pipeable Collection reduce
 */

export function reduce(b, f) {
  return self => {
    let res = b;
    const iterator = self[Symbol.iterator](); // eslint-disable-next-line no-constant-condition

    while (true) {
      const result = iterator.next();

      if (result.done) {
        break;
      }

      res = f(res, result.value);
    }

    return res;
  };
}
/**
 * Loops over the iterator accumulating a result using the provided function and AssociativeIdentity giving access to the element index
 *
 * @tsplus static Collection.Aspects foldMapWithIndex
 * @tsplus pipeable Collection foldMapWithIndex
 */

export function foldMapWithIndex(M, f) {
  return self => {
    let res = M.identity;
    let n = -1;
    const iterator = self[Symbol.iterator](); // eslint-disable-next-line no-constant-condition

    while (true) {
      const result = iterator.next();

      if (result.done) {
        break;
      }

      n += 1;
      res = M.combine(res, f(result.value, n));
    }

    return res;
  };
}
/**
 * Loops over the iterator accumulating a result using the provided function and AssociativeIdentity
 *
 * @tsplus static Collection.Aspects foldMap
 * @tsplus pipeable Collection foldMap
 */

export function foldMap(M, f) {
  return self => {
    let res = M.identity;
    const iterator = self[Symbol.iterator](); // eslint-disable-next-line no-constant-condition

    while (true) {
      const result = iterator.next();

      if (result.done) {
        break;
      }

      res = M.combine(res, f(result.value));
    }

    return res;
  };
}
/**
 * Concats iterators together
 *
 * @tsplus pipeable-operator Collection &
 * @tsplus static Collection.Aspects concat
 * @tsplus pipeable Collection concat
 */

function concat_1(that) {
  return self => {
    return {
      *[Symbol.iterator]() {
        for (const x of self) {
          yield x;
        }

        for (const x of that) {
          yield x;
        }
      }

    };
  };
}
/**
 * Concats iterators together that are strictly of the same type
 *
 * @tsplus pipeable-operator Collection +
 */


export function concatOperator(that) {
  return self => concat_1(that)(self);
}
/**
 * Prepends a value to an iterator
 *
 * @tsplus static Collection.Aspects prepend
 * @tsplus pipeable Collection prepend
 */

function prepend_1(that) {
  return self => {
    return {
      *[Symbol.iterator]() {
        yield that;

        for (const x of self) {
          yield x;
        }
      }

    };
  };
}
/**
 * Prepends a value to an iterator
 *
 * @tsplus pipeable-operator Collection >
 */


function prependOperator_1(self) {
  return a => prepend_1(a)(self);
}
/**
 * Prepends a value to an iterator of the same type
 *
 * @tsplus pipeable-operator Collection + 1.0
 */


export function prependOperatorStrict(self) {
  return a => prependOperator_1(self)(a);
}
/**
 * Appends a value to an iterator.
 *
 * @tsplus pipeable-operator Collection <
 * @tsplus static Collection.Aspects append
 * @tsplus pipeable Collection append
 */

function append_1(that) {
  return self => {
    return {
      *[Symbol.iterator]() {
        for (const x of self) {
          yield x;
        }

        yield that;
      }

    };
  };
}
/**
 * Appends a value to an iterator of the same type
 *
 * @tsplus pipeable-operator Collection + 1.0
 */


export function appendOperatorStrict(a) {
  return self => append_1(a)(self);
}
/**
 * Flattens nested iterators
 *
 * @tsplus getter Collection flatten
 */

export function flatten(self) {
  return flatMap_1(tsplus_module_1.identity)(self);
}
/**
 * Infinite sequence produced by repeated application of f to a
 *
 * @tsplus static Collection.Ops unfold
 */

export function unfold(a, f) {
  return {
    *[Symbol.iterator]() {
      yield a;
      let current = a;

      while (true) {
        current = f(a);
        yield current;
      }
    }

  };
}
/**
 * Compares each element of the iterators using the provided function
 *
 * @tsplus static Collection.Aspects equalsWith
 * @tsplus pipeable Collection equalsWith
 */

function equalsWith_1(that, f) {
  return self => {
    const leftIt = self[Symbol.iterator]();
    const rightIt = that[Symbol.iterator](); // eslint-disable-next-line no-constant-condition

    while (1) {
      const lnext = leftIt.next();
      const rnext = rightIt.next();

      if (lnext.done !== rnext.done) {
        return false;
      }

      if (lnext.done) {
        return true;
      }

      if (!f(lnext.value, rnext.value)) {
        return false;
      }
    }

    throw new Error("Bug");
  };
}

export function equals(that) {
  return self => equalsWith_1(that, tsplus_module_2.equals)(self);
}
/**
 * @tsplus static Collection.Ops make
 * @tsplus static Collection.Ops __call
 */

export function make(...as) {
  return as;
}
/**
 * @tsplus getter Collection toArray
 */

export function toArray(data) {
  return Array.from(data);
}
//# sourceMappingURL=functions.mjs.map