"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.MutableRandom = void 0;

var _Distribution = /*#__PURE__*/require("@tsplus/stdlib/io/Random/Distribution");

/**
 * Wrapper around an instance of a `RandomGenerator` random number generator
 * offering a simpler interface to deal with random with impure patterns.
 */
class MutableRandom {
  /**
   * Create a mutable random number generator.
   *
   * @param internalRng - Immutable random number generator
   */
  constructor(internalRng) {
    this.internalRng = internalRng;
  }
  /**
   * Clone the random number generator.
   */


  clone() {
    return new MutableRandom(this.internalRng);
  }
  /**
   * Generate an integer having `bits` random bits.
   *
   * @param bits - Number of bits to generate
   */


  next(bits) {
    return this.uniformIn(0, (1 << bits) - 1);
  }
  /**
   * Generate a random boolean.
   */


  nextBoolean() {
    return this.uniformIn(0, 1) === 1;
  }

  nextInt(min, max) {
    return this.uniformIn(min == null ? MutableRandom.MIN_INT : min, max == null ? MutableRandom.MAX_INT : max);
  }
  /**
   * Generate a random `bigint` between min (inclusive) and max (inclusive).
   *
   * @param min - Minimal `bigint` value
   * @param max - Maximal `bigint` value
   */


  nextBigInt(min, max) {
    return (0, _Distribution.uniformBigIntDistribution)(min, max, this.internalRng);
  }
  /**
   * Generate a random `ArrayInt` between min (inclusive) and max (inclusive).
   *
   * @param min - Minimal `ArrayInt` value
   * @param max - Maximal `ArrayInt` value
   */


  nextArrayInt(min, max) {
    return (0, _Distribution.uniformArrayIntDistribution)(min, max, this.internalRng);
  }
  /**
   * Generate a random floating point number between 0.0 (inclusive) and 1.0
   * (exclusive).
   */


  nextDouble() {
    const a = this.next(26);
    const b = this.next(27);
    return (a * MutableRandom.DBL_FACTOR + b) * MutableRandom.DBL_DIVISOR;
  }

  setSeed(seed) {
    this.internalRng.setSeed(seed);
  }

  uniformIn(rangeMin, rangeMax) {
    return (0, _Distribution.uniformIntDistribution)(rangeMin, rangeMax, this.internalRng);
  }

}

exports.MutableRandom = MutableRandom;
MutableRandom.MIN_INT = 0x80000000 | 0;
MutableRandom.MAX_INT = 0x7fffffff | 0;
MutableRandom.DBL_FACTOR = /*#__PURE__*/Math.pow(2, 27);
MutableRandom.DBL_DIVISOR = /*#__PURE__*/Math.pow(2, -53);
//# sourceMappingURL=MutableRandom.js.map